#include "runsettings.h"
#include <ctype.h>
#include <math.h>

#ifdef WINDOWS
#include <strstrea.h>
#else
#include <strstream.h>
#endif 


const char* RunSettings::paramTypes[] 
  = { "int", "algorithm", "double", "double[]", "string", "boolean", 
      "model", "in-format"};

const char* RunSettings::algorithms[] = { "global", "local" };

const char* RunSettings::models[] = { "HKY85", "F84", "TN93" };

const char* RunSettings::initialTreeTypes[] = { "random", "upgma",
						"neighbor-joining",
						"bambe", "newick" };

/* Names, types, and default values of parameters in the rc file.
   The parameters will be processed in the order that they are
   listed in this array. */

const char* RunSettings::inputParams[43][3] = {
  "seed",			"int",	       "194024933",
  "burn",			"int",	       "1000",
  "burn-algorithm",		"algorithm",   "global",
  "main-algorithm",		"algorithm",   "local",
  "cycles",			"int",	       "6000",
  "sample-interval",		"int",	       "10",
  "parameter-update-interval",	"int",	       "1",
  "update-kappa",           	"boolean",     "true",
  "update-theta",              	"boolean",     "true",
  "update-pi",                 	"boolean",     "true",
  "update-ttp",                	"boolean",     "true",
  "update-gamma",              	"boolean",     "true",
  "tune-interval",              "int",         "200",
  "window-interval",		"int",         "200",
  "molecular-clock",		"boolean",     "true",
  "likelihood-model",		"model",       "HKY85",
  "category-list",		"string",      "1*",
  "single-kappa",		"boolean",     "false",
  "initial-kappa",		"double[]",    "2.0",
  "initial-theta",		"double[]",    "1.0",
  "estimate-pi",		"boolean",     "true",
  "initial-pia",		"double[]",    "0.25",
  "initial-pig",		"double[]",    "0.25",
  "initial-pic",		"double[]",    "0.25",
  "initial-pit",		"double[]",    "0.25",
  "initial-ttp",		"double[]",    "1.0",
  "initial-gamma",		"double[]",    "1.0",
  "data-file",			"string",      "infile",
  "outgroup",                  	"int",         "1",
  "global-tune",		"double",      "0.01",
  "kappa-tune",			"double",      "0.1",
  "theta-tune",               	"double",      "2000.0",
  "pi-tune",                  	"double",      "2000.0",
  "ttp-tune",			"double",      "0.1",
  "gamma-tune",			"double",      "0.1",
  "local-tune",			"double",      "0.19",
  "use-beta",			"boolean",     "false",
  "beta-tune",			"double",      "10.0",
  "max-initial-tree-height",	"double",      ".1",
  "file-root",			"string",      "run1",
  "initial-tree-type",		"in-format",   "random",
  "tree-file",			"string",      "treefile",
  "newick-format",		"boolean",     "true"
};

const int RunSettings::numParamTypes = sizeof(paramTypes)
     / sizeof(paramTypes[0]);
const int RunSettings::numAlgorithms = sizeof(algorithms)
     / sizeof(algorithms[0]);
const int RunSettings::numModels = sizeof(models)/sizeof(models[0]);
const int RunSettings::numInitialTreeTypes = sizeof(initialTreeTypes)
  /sizeof(initialTreeTypes[0]);
const int RunSettings::numInputParams = sizeof(inputParams)
     / sizeof(inputParams[0]);

int RunSettings::findInputParam(const char *s)
{
  /* Returns the number of a parameter given the name s of a valid parameter.
     Otherwise returns the total number of parameters. */
  int i;

  for(i=0;i<numInputParams;i++)
    if(strcmp(s,inputParams[i][0])==0)
      return i;
  return i;
}

int RunSettings::stringMember(const char *s, int n, const char *sl[])
{
  /* Returns the index of string s in the list of strings sl (of size n),
     if s is a member of that list. Returns n otherwise. */
  int i;

  for(i=0;i<n;i++)
    if(strcmp(s,sl[i])==0)
      return i;
  return i;
}

int RunSettings::stringPrefix(const char *s, int n, const char *sl[])
{
  /* Returns the index of a string in the list of strings sl (of size n),
     of which s is a prefix, if such a string exists. Returns n otherwise. */
  int i;

  for(i=0;i<n;i++)
    if(strncmp(s,sl[i],strlen(sl[i]))==0)
      return i;
  return i;
}

void RunSettings::printBadParam(char *param, int lineno)
{
  error << "'" << stripRightSpaces(param,param+strlen(param)) 
	<<"' (line " << lineno << ")." << endError;

}

void RunSettings::addCategory(int n, int *cats)
{
  int *i,*j;

  for(i=cats;i<cats+numTags && *i<=n;i++)
    if(*i == n)
      return;
  for(j=cats+numTags;j>i;j--)
    *j = *(j-1);
  *j = n;
  numTags++;
}

int RunSettings::getCategories(char **str, int *cats, char *param, 
				int lineno) 
{
  /* Parses a string representing a list of categories up to, but
     not including the final asterisk (if it exists).
     Checks for syntax errors and sets up the category array.
     str is the unscanned part of the category list string.
     cats is an array of categories (in order) seen so far.
     numTags is the number of distinct categories seen so far.
     Returns whether or not a category was read in from the string. */

  int n;

  while(isspace(**str))
    (*str)++;
  if(**str == '(') {
    (*str)++;
    if(!getCategories(str,cats,param,lineno)) {
      error << "Error: Empty group in category list ";
      printBadParam(param,lineno);
      quit(1);
    }
    while(isspace(**str))
      (*str)++;
    if(**str == '*') {
      error << "Error: Asterisk must be the last symbol in the category list ";
      printBadParam(param,lineno);
      quit(1);
    }
    if(**str != ')') {
      error << "Error: Mismatched parentheses in category list ";
      printBadParam(param,lineno);
      quit(1);
    }
    (*str)++;
  }
  else if(isdigit(**str)) {
    n = **str - '0';
    for((*str)++;isdigit(**str);(*str)++)
      n = 10 * n + **str - '0';
    if(n == 0) {
      error << "Error: Category must be a positive number in category list ";
      printBadParam(param,lineno);
      quit(1);
    }
    else if(n > MAX_CATEGORIES) {
      error << "Error: Category must be less than "  << MAX_CATEGORIES
	    << " in category list ";
      printBadParam(param,lineno);
      quit(1);
    }
    addCategory(n,cats);
  }
  else 
    return 0;
  while(isspace(**str))
    (*str)++;
  if(**str == '^') {
    (*str)++;
    while(isspace(**str))
      (*str)++;
    if(!isdigit(**str)) {
      error << "Error: Repeat count must be an integer in the category list ";
      printBadParam(param,lineno);
      quit(1);
    }
    n = **str - '0';
    for((*str)++;isdigit(**str);(*str)++)
      n = 10 * n + **str - '0';
    if(n == 0) {
      error << "Error: Repeat count may not be zero in the category list ";
      printBadParam(param,lineno);
      quit(1);
    }
    while(isspace(**str))
      (*str)++;
  }
  if(**str == ',') {
    (*str)++;
    getCategories(str,cats,param,lineno);
  }
  return 1;
}

int RunSettings::intInRange(int x, int positive, int i, int *seen)
{
  if(positive ? x > 0 : x >= 0)
    return 1;
  error << "Error: " << inputParams[i][0] << "(" << x << ") must be "
	<< (positive ? "positive" : "non-negative") << " (line " << seen[i] 
	<< ")" << endError;
  return 0;
}

int RunSettings::doubleInRange(double x, int positive, int i, int *seen)
{
  if(positive ? x > 0 : x >= 0)
    return 1;
  error << "Error: " << inputParams[i][0] << "(" << x << ") must be "
	<< (positive ? "positive" : "non-negative") << " (line " << seen[i] 
	<< ")" << endError;
  return 0;
}

void RunSettings::parseParamList(int numvals, double *input, 
				 Vector<double>& valarray, int *given, int i, 
				 int *seen, double *ardoubleval,
				 int single, int ignore, int positive)
{ 

  const double tolerance = 1.0e-14;
  int j,k,warned = 0;

  if(ignore) {
    if(seen[i] > 0)
      error << "Warning: Values are to be estimated, not set in parameters "
	    << "file." << endl << "Line " << seen[i] << " ignored."
	    << endError;
    return;
  }
  valarray.setSize(numTags);
  if(numvals >= numCategories) {
    if(numvals > numCategories)
      error << "Warning: Too many values in " << inputParams[i][0] 
	    << " (line " << seen[i] << "). Extra values ignored." << endError;
    for(j=0;j<numCategories;j++)
      input[j] = ardoubleval[j];
    for(k=0;k<numTags;k++) {
      j = categories[k];
      if(!given[j-1]) {
	error << "Error: Value must be specified for " << inputParams[i][0] 
	      << " of category " << j << " (line " << seen[i] << ")." 
	      << endError;
	quit(1);
      }
      else if(!doubleInRange(ardoubleval[j-1],positive,i,seen)) 
	quit(1);
      else if(single && k > 0 && 
	      fabs(ardoubleval[j-1]-valarray[0])>tolerance) {
	if(!warned) {
	  error << "Warning: " << inputParams[i][0] << " values differ in "
		<< "single kappa mode. Only first value will be used (line " 
		<< seen[i] << ")." << endError;
	  warned = 1;
	}
	valarray[k] = valarray[0];
      }
      else
	valarray[k] = ardoubleval[j-1];
    }
  }
  else if(numvals == 1 && (single || seen[i]==0)) {
    if(!doubleInRange(ardoubleval[0],positive,i,seen)) 
      quit(1);
    for(k=0;k<numTags;k++)
      valarray[k] = ardoubleval[0];
    for(j=0;j<numCategories;j++)
      input[j] = ardoubleval[0];
  }
  else {
    error << "Error: Too few values in " << inputParams[i][0] << " (line "
	  << seen[i] << ")." << endError;
    quit(1);
  }
}

void RunSettings::fndString(char *v, int n, const char* vl[], 
			    int& okay, int& intval)
{
  stripRightSpaces(v,v+strlen(v));
  okay = ((intval=stringMember(v,n,vl)) < n);
}

void RunSettings::checkParamUsed(int i, int seen, int paramUpdated, 
				 const char* const paramName, int* okModels)
{
  if(seen) {
    if(!okModels[model])
      error << "Warning: parameter (" << inputParams[i][0] 
	    << ") not used in model " << models[model] << " (line "
	    << seen << ")." << endError;
    else if(!paramUpdated)
	error << "Warning: parameter (" << inputParams[i][0] 
	      << ") not used since " << paramName << " is not updated " 
	      << "(line " << seen << ")." << endError;
      else if(paramUpdateInterval == 0)
	error << "Warning: parameter (" << inputParams[i][0] 
	      << ") not used since parameters are never updated " << "(line "
	      << seen << ")." << endError;
  }
}

void RunSettings::resetOkModels(int* okModels)
{
  for(int j=0;j<numModels;j++)
    okModels[j] = 0;
}

RunSettings::RunSettings(istream& f)
{
  /* Reads in and processes the parameters from file f. Sets
     the corresponding object variables. A parameter not listed in the
     file is set to its default value. The parameters are processed
     in the order given by the parameter array. */

  char *pos,*eqpos,*pos1,*pos2,buf[MAX_LINE];
  char valx[MAX_LINE];
  int i,n,okay;
  int intval;
  double doubleval, ardoubleval[MAX_CATEGORIES];
  char *stringval;
  int elementGiven[MAX_CATEGORIES];
  int okModels[numModels];


  /* Set the value of each parameter to its default value and mark
     each as unseen. */

  char val[numInputParams][MAX_LINE];
  int seen[numInputParams];
  for(int i1=0;i1<numInputParams;i1++) {
    strcpy(val[i1],inputParams[i1][2]);
    seen[i1] = 0;
  }

  /* Read in each line of the file. Parse the parameter name and its
     value. */
  i = 0;
  while(f.getline(buf,MAX_LINE)) {
    i++;
    if((pos=strchr(buf,'#'))!=0)
      *pos = '\0';
    pos = buf+strspn(buf," \t\n");
    if(*pos!='\0')
      if((eqpos=strchr(pos,'='))!=0) {
	stripRightSpaces(pos,eqpos);
	if((n=findInputParam(pos)) < numInputParams) {
	  if(seen[n])
	    error << "Warning: parameter (" << inputParams[n][0] 
		  << ") multiply defined (line " << seen[n] << " and line "
		  << i << ")." << endError;
	  eqpos = eqpos+1+strspn(eqpos+1," \t\n");
	  seen[n] = i;
	  strcpy(val[n],eqpos);
	}
	else {
	  error << "Error: Unknown parameter name (" << pos << ") (line " 
		<< i << ")." << endError;
	  quit(1);
	}
      }
      else {
	error << "Error: Illegal syntax. Missing '=' (line " << i << ")." 
	      << endError;
	quit(1);
      }
  }
  /* Process the parameters in order, checking the types and numbers 
     of the values. */
  for(i=0;i<numInputParams;i++) {
    switch (stringMember(inputParams[i][1],numParamTypes,paramTypes)) {
    case INT: okay = (((istrstream(val[i],MAX_LINE))>>intval)!=NULL); break;
    case ALGORITHM: fndString(val[i],numAlgorithms,algorithms,okay,intval);
      break;
    case MODEL: fndString(val[i],numModels,models,okay,intval); break;
    case IN_FORMAT: 
      fndString(val[i],numInitialTreeTypes,initialTreeTypes,okay,intval);
      break;
    case DOUBLE: okay = (((istrstream(val[i],MAX_LINE))>>doubleval)!=NULL);
      break;
    case STRING: 
      stripRightSpaces(val[i],val[i]+strlen(val[i]));
      stringval = val[i];
      okay = 1;
      break;
    case DOUBLEARRAY:
      okay = 1;
      strcpy(valx,val[i]);
      for(pos1=valx,intval=0;*pos1;pos1=pos2,intval++) {
	if(intval>=MAX_CATEGORIES) {
	  error << "Warning: number of values exceeds maximum possible '"
	    << stripRightSpaces(val[i],val[i]+strlen(val[i])) << "' (line "
	    << seen[i] << ")." << endError;
	  break;
	}
	for(pos2=pos1;*pos2;pos2++)
	  if(*pos2 == ',') {
	    *pos2++ = '\0';
	    break;
	  }
	if(*pos1 == '\0') {
	  elementGiven[intval] = 0;
	  ardoubleval[intval] = 0;
	}
	else if(((istrstream(pos1,MAX_LINE))>>ardoubleval[intval])!=NULL)
	  elementGiven[intval] = 1;
	else {
	  ardoubleval[intval] = 0;
	  okay = 0;
	}
      }
      break;
    case BOOLEAN:
      strcpy(val[i],val[i]+strspn(buf," \t"));
      stripRightSpaces(val[i],val[i]+strlen(val[i]));
      if(strcmp(val[i],"true")==0) {
	intval = 1;
	okay = 1;
      }
      else if(strcmp(val[i],"false")==0) {
	intval = 0;
	okay = 1;
      }
      else 
	okay = 0;
      break;
    default: error << "Internal error: Unknown parameter type (" 
		   << inputParams[i][1] << ")." << endError;
      quit(1);
    }
    if(!okay) {
      error << "Error: Could not understand parameter value '"
	    << stripRightSpaces(val[i],val[i]+strlen(val[i])) << "' (line "
	    << seen[i] << ")." << endError;
      quit(1);
    }
    /* Process the parameter values in order, setting the appropriate
       global variables. */
    switch (i) {
    case SEED: seed = intval % (1UL << 31); 
      if( seed % 2 == 0 ) {  
	seed++;
	error << "Warning: changing initial seed to be odd number" << endError;
      }
      break;
    case BURN: 
      if(!intInRange(intval,0,i,seen))
	quit(1);
      burn = intval; 
      break;
    case BURN_ALGORITHM: burnAlgorithm = intval; break;
    case MAIN_ALGORITHM: mainAlgorithm = intval; break;
    case CYCLES: 
      if(!intInRange(intval,0,i,seen))
	quit(1);
      numCycles = intval;
      break;
    case SAMPLE_INTERVAL: 
      if(!intInRange(intval,1,i,seen))
	quit(1);
      sampleInterval = intval;
      break;
    case PARAMETER_UPDATE_INTERVAL: 
      if(!intInRange(intval,0,i,seen))
	quit(1);
      paramUpdateInterval = intval; 
      break;
    case UPDATE_KAPPA: updateKappa = intval; break;
    case UPDATE_THETA: updateTheta = intval; break;
    case UPDATE_PI: updatePi = intval; break;
    case UPDATE_TTP: updateTtp = intval; break;
    case UPDATE_GAMMA: updateGamma = intval; break;
    case TUNE_INTERVAL: 
      if(!intInRange(intval,1,i,seen))
	quit(1);
      tuneInterval = intval; 
      break;
    case WINDOW_INTERVAL: 
      if(!intInRange(intval,1,i,seen))
	quit(1);
      windowInterval = intval; 
      break;
    case MOLECULAR_CLOCK: mclock = intval; break;
    case LIKELIHOOD_MODEL: model = intval; break; 
    case CATEGORY_LIST:
      strcpy(categoryList,stringval);
      numTags = 0;
      if(!getCategories(&stringval,categories,val[i],seen[i])) {
	error << "Error: Empty category list ";
	printBadParam(val[i],seen[i]);
	quit(1);
      }
      if(*stringval == '*') {
	(stringval)++;
	while(isspace(*stringval))
	  (stringval)++;
      }
      if(*stringval != '\0') {
	error << "Error: Category list ends badly ";
	printBadParam(val[i],seen[i]);
	quit(1);
      }
      numCategories = categories[numTags-1];
      break;
    case SINGLE_KAPPA: singleKappa = intval; break;
    case INITIAL_KAPPA: 
      parseParamList(intval,inputKappa,initialKappa,elementGiven,
		      i,seen,ardoubleval,singleKappa,0,1);
      break;
    case INITIAL_THETA:
      parseParamList(intval,inputTheta,initialTheta,elementGiven,
		       i,seen,ardoubleval,0,0,0);
      break;
    case ESTIMATE_PI: estimatePi = intval; break;
    case INITIAL_PIA: 
      parseParamList(intval,inputPi[0],initialPi[0],elementGiven,
		       i,seen,ardoubleval,0,estimatePi,1);
      break;
    case INITIAL_PIG:
      parseParamList(intval,inputPi[1],initialPi[1],elementGiven,
		       i,seen,ardoubleval,0,estimatePi,1);
      break;
    case INITIAL_PIC:
      parseParamList(intval,inputPi[2],initialPi[2],elementGiven,
		       i,seen,ardoubleval,0,estimatePi,1);
      break;
    case INITIAL_PIT:
      parseParamList(intval,inputPi[3],initialPi[3],elementGiven,
		       i,seen,ardoubleval,0,estimatePi,1);
      break;
    case INITIAL_TTP:
      if(model==TN93) 
	parseParamList(intval,inputTtp,initialTtp,elementGiven,
			i,seen,ardoubleval,singleKappa,0,1);
      else if(seen[INITIAL_TTP])
	error << "Warning: parameter (" << inputParams[i][0] 
	      << ") not used in model " << models[model] << " (line " 
	      << seen[i] << ")." << endError;
      break;
    case INITIAL_GAMMA: 
      if(model==TN93) 
	parseParamList(intval,inputGamma,initialGamma,elementGiven,
			i,seen,ardoubleval,singleKappa,0,1);
      else if(seen[INITIAL_GAMMA])
	error << "Warning: parameter (" << inputParams[i][0] 
	      << ") not used in model " << models[model] << " (line "
	      << seen[i] << ")." << endError;
      break;
    case DATA_FILE: strcpy(dataFile,stringval); break;
    case OUTGROUP: 
      if(seen[i] && mclock)
	error << "Warning: parameter (" << inputParams[i][0] 
	      << ") not used with molecular clock " << "(line "
	      << seen[i] << ")." << endError;
      if(!intInRange(intval,1,i,seen))
	quit(1);
      outgroup = intval; 
      break;
    case GLOBAL_TUNE: 
      if(seen[GLOBAL_TUNE] && mainAlgorithm != GLOBAL && 
	 burnAlgorithm != GLOBAL)
	error << "Warning: parameter (" << inputParams[i][0] 
	      << ") not used in either algorithm " << "(line "
	      << seen[i] << ")." << endError;
      if(!doubleInRange(doubleval,1,i,seen))
	quit(1);
      valleyWin  = doubleval; 
      break;
    case KAPPA_TUNE: 
      resetOkModels(okModels);
      okModels[HKY85] = okModels[F84] = 1;
      checkParamUsed(i,seen[i],updateKappa,"kappa",okModels);
      if(!doubleInRange(doubleval,0,i,seen))
	quit(1);
      kappaWin = doubleval; 
      break;
    case THETA_TUNE:
      resetOkModels(okModels);
      okModels[HKY85] = okModels[F84] = okModels[TN93] = 1;
      checkParamUsed(i,seen[i],updateTheta,"theta",okModels);
      if(!doubleInRange(doubleval,0,i,seen))
	quit(1);
      thetaConst  = doubleval; 
      break;
    case PI_TUNE:
      resetOkModels(okModels);
      okModels[HKY85] = okModels[F84] = okModels[TN93] = 1;
      checkParamUsed(i,seen[i],updatePi,"pi",okModels);
      if(!doubleInRange(doubleval,0,i,seen))
	quit(1);
      piConst  = doubleval; 
      break;
    case TTP_TUNE:
      resetOkModels(okModels);
      okModels[TN93] = 1;
      checkParamUsed(i,seen[i],updateTtp,"ttp",okModels);
      if(!doubleInRange(doubleval,0,i,seen))
	quit(1);
      ttpWin  = doubleval; 
      break;
    case GAMMA_TUNE:
      resetOkModels(okModels);
      okModels[TN93] = 1;
      checkParamUsed(i,seen[i],updateGamma,"gamma",okModels);
      if(!doubleInRange(doubleval,0,i,seen))
	quit(1);
      gammaWin  = doubleval; 
      break;
    case LOCAL_TUNE:
      if(seen[LOCAL_TUNE] && mainAlgorithm != LOCAL && burnAlgorithm != LOCAL)
	error << "Warning: parameter (" << inputParams[i][0] 
	      << ") not used in either algorithm " << "(line "
	      << seen[i] << ")." << endError;
      if(!doubleInRange(doubleval,1,i,seen))
	quit(1);
      localTune  = doubleval; 
      break;
    case USE_BETA: useBeta = intval; break;
    case BETA_TUNE: 
      if(seen[BETA_TUNE] && !useBeta)
	error << "Warning: parameter (" << inputParams[i][0] 
	      << ") not used in either algorithm " << "(line "
	      << seen[i] << ")." << endError;
      if(!doubleInRange(doubleval,1,i,seen))
	quit(1);
      betaTune = doubleval; 
      break;
    case MAX_INITIAL_TREE_HEIGHT: 
      if(!doubleInRange(doubleval,1,i,seen))
	quit(1);
      maxDepth = doubleval; 
      break;
    case FILE_ROOT: strcpy(fileRoot,stringval); break;
    case INITIAL_TREE_TYPE: 
      if(intval == NEIGH && mclock) {
	error << "Error: Neighbor joining cannot be used with the molecular "
	      << "clock " << "(line " << seen[i] << ")." << endError;
	quit(1);
      }
      initialTreeType = intval; 
      break;
    case TREE_FILE: 
      if(seen[TREE_FILE] && initialTreeType != BAMBE && 
	 initialTreeType != NEWICK)
	error << "Warning: parameter (" << inputParams[i][0] 
	      << ") not used for " << initialTreeTypes[initialTreeType]
	      << " tree creation method " << "(line " << seen[i] << ")." 
	      << endError;
      strcpy(treeFile,stringval);
      break;
    case NEWICK_FORMAT: newickFormat = intval; break;
    default: error << "Internal error: Unknown parameter number " << i << "."
		   << endError;
    }
  }
}

void RunSettings::print(ostream& f) const
{
  /* Prints out the parameters and their values. */
  f << "Run Settings:" << endl << endl;
  int i;
  f << inputParams[SEED][0] << "=" << seed << endl;
  f << inputParams[BURN][0] << "=" << burn << endl;
  f << inputParams[BURN_ALGORITHM][0] << "=" 
    << algorithms[burnAlgorithm] << endl;
  f << inputParams[MAIN_ALGORITHM][0] << "=" 
    << algorithms[mainAlgorithm] << endl;
  f << inputParams[CYCLES][0] << "=" << numCycles << endl;
  f << inputParams[SAMPLE_INTERVAL][0] << "=" << sampleInterval << endl;
  f << inputParams[PARAMETER_UPDATE_INTERVAL][0] << "=" 
    << paramUpdateInterval << endl;
  f << inputParams[UPDATE_KAPPA][0] << "="
    << (updateKappa?"true":"false") << endl;
  f << inputParams[UPDATE_THETA][0] << "="
    << (updateTheta?"true":"false") << endl;
  f << inputParams[UPDATE_PI][0] << "=" 
    << (updatePi?"true":"false") << endl;
  f << inputParams[UPDATE_TTP][0] << "=" 
    << (updateTtp?"true":"false") << endl;
  f << inputParams[UPDATE_GAMMA][0] << "=" 
    << (updateGamma?"true":"false") << endl;
  f << inputParams[TUNE_INTERVAL][0] << "=" << tuneInterval << endl;
  f << inputParams[WINDOW_INTERVAL][0] << "=" << windowInterval << endl;
  f << inputParams[MOLECULAR_CLOCK][0] << "=" 
    << (mclock?"true":"false") << endl;
  f << inputParams[LIKELIHOOD_MODEL][0] << "=" << models[model] << endl;
  f << inputParams[CATEGORY_LIST][0] << "=" << categoryList << endl;
  f << inputParams[SINGLE_KAPPA][0] << "=" << 
	  (singleKappa?"true":"false") << endl;
  f << inputParams[INITIAL_KAPPA][0] << "=" << inputKappa[0];
  if(!singleKappa)
    for(i=1;i<numCategories;i++)
      f << "," << inputKappa[i];
  f << endl;
  f << inputParams[INITIAL_THETA][0] << "=" << inputTheta[0];
  for(i=1;i<numCategories;i++)
    f << "," << inputTheta[i];
  f << endl;
  f << inputParams[ESTIMATE_PI][0] << "=" 
  << (estimatePi?"true":"false") << endl;
  if(!estimatePi) {
    f << inputParams[INITIAL_PIA][0] << "=" << inputPi[0][0];
    for(i=1;i<numCategories;i++)
      f << "," << inputPi[0][i];
    f << endl;
    f << inputParams[INITIAL_PIG][0] << "=" << inputPi[1][0];
    for(i=1;i<numCategories;i++)
      f << "," << inputPi[1][i];
    f << endl;
    f << inputParams[INITIAL_PIC][0] << "=" << inputPi[2][0];
    for(i=1;i<numCategories;i++)
      f << "," << inputPi[2][i];
    f << endl;
    f << inputParams[INITIAL_PIT][0] << "=" << inputPi[3][0];
    for(i=1;i<numCategories;i++)
      f << "," << inputPi[3][i];
    f << endl;
  }
  if(model==TN93) {
    f << inputParams[INITIAL_TTP][0] << "=" << inputTtp[0];
    if(!singleKappa)
      for(i=1;i<numCategories;i++)
	f << "," << inputTtp[i];
    f << endl;
    f << inputParams[INITIAL_GAMMA][0] << "=" << inputGamma[0];
    if(!singleKappa)
      for(i=1;i<numCategories;i++)
	f << "," << inputGamma[i];
    f << endl;
  }
  f << inputParams[DATA_FILE][0] << "=" << dataFile << endl;
  f << inputParams[OUTGROUP][0] << "=" << outgroup << endl;
  if(mainAlgorithm == GLOBAL || burnAlgorithm == GLOBAL)
    f << inputParams[GLOBAL_TUNE][0] << "=" << valleyWin << endl;
  if(model!=TN93)
    f << inputParams[KAPPA_TUNE][0] << "=" << kappaWin << endl;
  f << inputParams[THETA_TUNE][0] << "=" << thetaConst << endl;
  f << inputParams[PI_TUNE][0] << "=" << piConst << endl;  
  if(model==TN93)
    f << inputParams[TTP_TUNE][0] << "=" << ttpWin << endl;  
  if(model==TN93)
    f << inputParams[GAMMA_TUNE][0] << "=" << gammaWin << endl;  
  if(mainAlgorithm == LOCAL || burnAlgorithm == LOCAL)
    f << inputParams[LOCAL_TUNE][0] << "=" << localTune << endl;  
  f << inputParams[USE_BETA][0] << "=" << 
	  (useBeta?"true":"false") << endl;
  if(useBeta)
    f << inputParams[BETA_TUNE][0] << "=" << betaTune << endl;  
  f << inputParams[MAX_INITIAL_TREE_HEIGHT][0] << "=" << maxDepth << endl;
  f << inputParams[FILE_ROOT][0] << "=" << fileRoot << endl;
  f << inputParams[INITIAL_TREE_TYPE][0] << "=" 
    << initialTreeTypes[initialTreeType] << endl;
  if(initialTreeType == BAMBE || initialTreeType == NEWICK)
    f << inputParams[TREE_FILE][0] << "=" << treeFile << endl;
  f << inputParams[NEWICK_FORMAT][0] << "=" 
  << (newickFormat?"true":"false") << endl;
}
