#ifndef RUNSETTINGSHDR
#define RUNSETTINGSHDR

#include "util.h"
#include <iostream.h>

#include <string.h>

#include <fstream.h>

class RunSettings {
public:
  enum { GLOBAL, LOCAL }; 			// Tree deformation algorithms.
  enum { HKY85, F84, TN93 }; 			// Probability models. 
  enum { RANDOM, UPGMA, NEIGH, BAMBE, NEWICK }; // Tree creation routines.
  RunSettings(istream&);
  void print(ostream&) const;
  uint32 getSeed() const { return seed; }
  ifstream& getFlogp() { return flogp; }
  ifstream& getFkappa() { return fkappa; }
  ifstream& getFtop() { return ftop; }
  ifstream& getFout() { return fout;}
  ifstream& getFtree() { return ftree;}
  int getNumCycles() const { return numCycles;}
  int getSampleInterval() const { return sampleInterval;}
  int getParamUpdateInterval() const { return paramUpdateInterval;}
  int getWindowInterval() const { return windowInterval;}
  int getBurn() const { return burn;}
  int getUpdateKappa() const { return updateKappa;}
  int getUpdateTheta() const { return updateTheta;}
  int getUpdatePi() const { return updatePi;}
  int getUpdateTtp() const { return updateTtp;}
  int getUpdateGamma() const { return updateGamma;}
  int getTuneInterval() const { return tuneInterval;}
  double getInitialKappa(int i) const { return initialKappa[i];}
  double getInitialTheta(int i) const { return initialTheta[i];}
  const Vector<double> getInitialPi(int i) const { return initialPi[i];}
  int getEstimatePi() const { return estimatePi;}
  double getInitialTtp(int i) const { return initialTtp[i];}
  double getInitialGamma(int i) const { return initialGamma[i];} 
  double getValleyWin() const { return valleyWin;}
  double getKappaWin() const { return kappaWin;}
  double getMaxDepth() const { return maxDepth;}
  const char* getDataFile() const { return dataFile;}
  const char* getTreeFile() const { return treeFile;}
  const char* getFileRoot() const { return fileRoot;}
  int getNumTags() const { return numTags;}
  int getNumCategories() const { return numCategories;}
  int getBurnAlgorithm() const { return burnAlgorithm;}
  int getMainAlgorithm() const { return mainAlgorithm;}
  int getSingleKappa() const { return singleKappa;}
  int getMclock() const { return mclock;}
  int getModel() const { return model;}
  int getOutGroup() const { return outgroup;}
  const char* getCategoryList() const { return categoryList; }
  int getCategories(int i) const { return categories[i]; }
  int getInitialTreeType() const { return initialTreeType; }
  int getNewickFormat() const { return newickFormat; }
  double getThetaConst() const { return thetaConst; }
  double getPiConst() const { return piConst; }
  double getTtpWin() const { return ttpWin; }
  double getGammaWin() const { return gammaWin; }
  double getLocalTune() const { return localTune; }
  int getUseBeta() const { return useBeta; }
  double getBetaTune() const { return betaTune; }
  void scaleInitialTheta(double denom) {
    for(int i=0;i<numTags;i++)
      initialTheta[i] /= denom; }
  void halveValleyWin() { valleyWin /= 2.0; }
  
 private:
  uint32 seed;			// odd unsigned int between 0 and
  				// (4294967295 = 2^32 - 1)
  ifstream flogp;		// log likelihood output file
  ifstream fkappa;		// parameter output file
  ifstream ftop;		// tree topology output file
  ifstream fout;		// summary statistics and run settings output
  ifstream ftree;		// final tree output file
  int numCycles;		// number of cycles after burn in
  int sampleInterval;		// interval for storing the tree
  int paramUpdateInterval;	// interval for updating parameters
  int windowInterval;		// interval for printing to the terminal
  int burn;			// the number of cycles during burn in
  int updateKappa;		// whether or not to update kappa
  int updateTheta;		// whether or not to update theta
  int updatePi;			// whether or not to update pi
  int updateTtp;		// whether or not to update ttp
  int updateGamma;		// whether or not to update gamma
  int tuneInterval;		// interval for updating global tune in burn in
  int estimatePi;		// estimate pi from base frequencies
  Vector<double> initialKappa;  // the inital value for kappa
  Vector<double> initialTheta;  // the inital value for theta
  Vector<double> initialPi[NUM_BASES]; // the inital value for pi
  Vector<double> initialTtp;    // the inital value for ttp
  Vector<double> initialGamma;  // the inital value for gamma
  double valleyWin;		// the size of half the window for
  				// updating valley depths.
  double kappaWin;		// the size of half the window for updating 
				// kappa.
  double maxDepth;		// the maximum size of a valley depth.
  				// for the initial tree.
  char dataFile[81];		// the name of the file where the data is 
  				// stored.
  char treeFile[81];		// file containing initial tree
  char fileRoot[81];		// prefix for names of output files
  int numTags;			// number of tags (used internally)
  int numCategories;		// number of categories (used externally)
  int burnAlgorithm;		// deformation algorithm used during burn in
  int mainAlgorithm;		// deformation algorithm used after burn in
  int singleKappa;		// whether or not all categories use same kappa
  int mclock;			// whether or not molecular clock is used
  int model;			// likelihood model
  int outgroup;			// outgroup for non-clock case
  char categoryList[MAX_LINE];  // list of categories
  int categories[MAX_CATEGORIES];// mapping for tag number to category number
  int initialTreeType;		// source of the input tree
  int newickFormat;		// whether or not to read and print in newick
				// format
  double thetaConst;		// tuning parameter for updating theta
  double piConst;		// tuning parameter for updating pi
  double ttpWin;		// tuning parameter for updating ttp
  double gammaWin;		// tuning parameter for updating gamma
  double localTune;		// tuning parameter for stretching factor in
				// local deformation algorithm
  int useBeta;			// whether or not to use the beta distribution
				// during local deformation
  double betaTune;		// tuning parameter for updating theta

  static const char *paramTypes[];
  static const int numParamTypes;
  static const char *algorithms[];
  static const int numAlgorithms;
  static const char *models[];
  static const int numModels;
  static const char *initialTreeTypes[];
  static const int numInitialTreeTypes;
  enum { INT, ALGORITHM, DOUBLE, DOUBLEARRAY, STRING, BOOLEAN, MODEL,
	 IN_FORMAT };
  static const char *inputParams[43][3];
  enum { SEED, BURN, BURN_ALGORITHM, MAIN_ALGORITHM, 
	 CYCLES, SAMPLE_INTERVAL, PARAMETER_UPDATE_INTERVAL,
	 UPDATE_KAPPA, UPDATE_THETA, UPDATE_PI, UPDATE_TTP, UPDATE_GAMMA,
	 TUNE_INTERVAL, WINDOW_INTERVAL, MOLECULAR_CLOCK, LIKELIHOOD_MODEL,
	 CATEGORY_LIST, SINGLE_KAPPA, INITIAL_KAPPA, INITIAL_THETA,
	 ESTIMATE_PI, INITIAL_PIA, INITIAL_PIG, INITIAL_PIC, 
	 INITIAL_PIT, INITIAL_TTP, INITIAL_GAMMA,
	 DATA_FILE, OUTGROUP, GLOBAL_TUNE, KAPPA_TUNE, THETA_TUNE, 
	 PI_TUNE, TTP_TUNE, GAMMA_TUNE, LOCAL_TUNE, USE_BETA, BETA_TUNE,
	 MAX_INITIAL_TREE_HEIGHT, FILE_ROOT, INITIAL_TREE_TYPE, TREE_FILE,
	 NEWICK_FORMAT };
  static const int numInputParams;

  double inputKappa[MAX_CATEGORIES];
  double inputTheta[MAX_CATEGORIES];
  double inputPi[NUM_BASES][MAX_CATEGORIES];
  double inputTtp[MAX_CATEGORIES];
  double inputGamma[MAX_CATEGORIES];

  int findInputParam(const char *);
  int stringMember(const char *, int, const char *[]);
  int stringPrefix(const char *, int, const char *[]);
  void fndString(char *, int, const char* [], int&, int&);
  void printBadParam(char *, int);
  void addCategory(int, int *);
  int getCategories(char **, int *, char *, int);
  int intInRange(int, int, int, int *);
  int doubleInRange(double, int, int, int *);
  void parseParamList(int, double *, Vector<double>&, int *, int, int *, 
			double *, int, int, int);
  void checkParamUsed(int,int,int,const char* const,int*);
  void resetOkModels(int*);
};

#endif
